@@ -2,30 +2,64 @@ from autograd import elementwise_grad, grad
 import autograd.numpy as np
 import copy
 from scipy.optimize import minimize
+import GPy
+import itertools
 
 from cwgp.kernel import RBF, OU, Matern32
 from cwgp.transformations import sal, sa, asinh, box_cox, inv_sal, inv_sa, inv_asinh, inv_box_cox
+from numba import jit
 
 
 class Phi():
     FN_BANK = {
-        "sal": {"fn": sal, "inv_fn": inv_sal, "par_len": 4},
-        "sa": {"fn": sa, "inv_fn": inv_sa, "par_len": 2},
-        "asinh": {"fn": asinh, "inv_fn": inv_asinh, "par_len": 4},
-        "box_cox": {"fn": box_cox, "inv_fn": inv_box_cox, "par_len": 1},
+        "sal": {
+            "fn": sal,
+            "inv_fn": inv_sal,
+            "par_len": 4,
+            "bounds": [
+                (-20,
+                 20)] * 4},
+        "sa": {
+            "fn": sa,
+            "inv_fn": inv_sa,
+            "par_len": 2,
+            "bounds": [
+                (-20,
+                 20)] * 2},
+        "asinh": {
+            "fn": asinh,
+            "inv_fn": inv_asinh,
+            "par_len": 4,
+            "bounds": [
+                (-20,
+                 20),
+                (1e-5,
+                 None),
+                (1e-5,
+                 None),
+                (-20,
+                 20)]},
+        "box_cox": {
+            "fn": box_cox,
+            "inv_fn": inv_box_cox,
+            "par_len": 1,
+            "bounds": [
+                (1e-5,
+                 None)]},
     }
 
     KERNEL_BANK = {
-        "OU": {"kern": OU, "params": 1, "init_scale": 1},
-        "RBF": {"kern": RBF, "params": 1, "init_scale": 10},
-        "Matern32": {"kern": Matern32, "params": 2, "init_scale": 10},
+        "OU": {"kern": GPy.kern.Exponential, "params": 2, "init_scale": 1},
+        "RBF": {"kern": GPy.kern.RBF, "params": 2, "init_scale": 1},
+        "Matern32": {"kern": GPy.kern.Matern32, "params": 2, "init_scale": 10},
     }
 
     def __init__(
             self,
             fn,
-            kernel="OU",
+            kernel="RBF",
             kernel_params_estimate=True,
+            ARD=False,
             **kwargs
     ):
         self.fn = [self.FN_BANK[f]["fn"]
@@ -35,8 +69,12 @@ class Phi():
         self.d_fn = [elementwise_grad(f, 1)
                      for f in self.fn]  # take derivative
         self.kernel = self.KERNEL_BANK[kernel]["kern"]
+        self.kernel_name = kernel
         self.kernel_params = self.KERNEL_BANK[kernel]["params"] if kernel_params_estimate else 0
         self.init_scale = self.KERNEL_BANK[kernel]["init_scale"]
+        self.ARD = ARD
+        self.bounds = list(itertools.chain.from_iterable(
+            [self.FN_BANK[f]["bounds"] for f in fn])) + [(None, None)] * self.kernel_params
 
     def comp_phi(self, par, y):
         assert len(par) >= sum(self.par_len), "Not enough parameters"
@@ -65,13 +103,15 @@ class Phi():
         t_t = np.transpose(t)
         mean_t = mf(t) if mf else np.zeros(t.shape)
         if self.kernel_params:
-            cov_xx = self.kernel(t, t_t, par[-self.kernel_params:])
+            cov_xx = self.kernel(
+                1, *par[-self.kernel_params:]).K(t.reshape(-1, 1), t.reshape(-1, 1))
         else:
-            cov_xx = self.kernel(t, t_t)
+            cov_xx = self.kernel(1).K(t.reshape(-1, 1), t.reshape(-1, 1))
+
         gaussian_params = 0.5 * \
             (t_t - np.transpose(mean_t)) @ np.linalg.inv(cov_xx) @ (t - mean_t)
         return np.ravel(0.5 * np.log(np.linalg.det(cov_xx)) +
-                        gaussian_params - sum(np.log(chain_d_sal)))
+                        gaussian_params - np.sum(np.log(chain_d_sal)))
 
     def reml(self, method):
         pass
@@ -80,28 +120,27 @@ class Phi():
             self,
             y,
             t,
-            mf=None,
-            method='l-bfgs-b',
-            loop=True,
-            verbose=False,
             **kwargs):
         # http://stackoverflow.com/questions/19843752/structure-of-inputs-to-scipy-minimize-function
+        mf = kwargs.get("mf", None)
         res = minimize
         res.success = False
-        if loop:
+        if kwargs.get("loop", True):
             while res.success == False:
                 try:
                     res = minimize(
                         self.likelihood,
-                        self.init_scale * np.random.rand(
-                            sum(self.par_len) +
-                            self.kernel_params),
+
+                        np.random.uniform(0, self.init_scale,
+                                          sum(self.par_len) + self.kernel_params),
                         args=(
                             y, t, mf
                         ),
-                        method=method)
+                        method=kwargs.get("method", "L-BFGS-B"),
+                        bounds=self.bounds,
+                    )
                 except Exception as e:
-                    if verbose:
+                    if kwargs.get("verbose", False):
                         print(e)
         self.res = res
         return res
